# -*- coding: utf-8 -*-

"""A single-point energy in LAMMPS"""

import logging
from pathlib import Path

from tabulate import tabulate

import lammps_step
import seamm
from seamm_util import units_class
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
printer = printing.getPrinter("lammps")


class Energy(seamm.Node):
    """Handle a singlepoint energy calculation in LAMMPS"""

    def __init__(
        self,
        flowchart=None,
        title="Energy",
        extension=None,
        logger=logger,
    ):
        """Initialize the node"""

        logger.debug("Creating Energy {}".format(self))

        super().__init__(
            flowchart=flowchart,
            title=title,
            extension=extension,
            logger=logger,
        )

        self._calculation = "energy"
        self._model = None
        self._metadata = lammps_step.metadata
        self.parameters = lammps_step.EnergyParameters()

        self.description = "A single point energy calculation"

    @property
    def model(self):
        """The name of the forcefield."""
        return self.parent.model

    @model.setter
    def model(self, value):
        self.parent.model = value

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}".format(".".join(str(e) for e in self._id), self.title)

    @property
    def results(self):
        """The storage for the results in the main LAMMPS step."""
        return self.flowchart.parent._results

    @property
    def version(self):
        """The semantic version of this module."""
        return lammps_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return lammps_step.__git_revision__

    def description_text(self, P=None):
        """Create the text description of what this step will do."""

        text = "Single-point energy calculation."

        return self.header + "\n" + __(text, indent=4 * " ").__str__()

    def analyze(self, indent="", data={}, table=None, output=[], **kwargs):
        """Parse the output and generating the text output and store the
        data in variables for other stages to access
        """
        if table is not None:
            text = ""
            tmp = tabulate(
                table,
                headers="keys",
                tablefmt="simple",
                disable_numparse=True,
                colalign=(
                    "center",
                    "decimal",
                    "center",
                    "decimal",
                    "left",
                    "decimal",
                    "decimal",
                    "decimal",
                ),
            )
            length = len(tmp.splitlines()[0])
            text += "\n"
            text += "Properties".center(length)
            text += "\n"
            text += tmp
            text += "\n"

            printer.normal(__(text, indent=8 * " ", wrap=False, dedent=False))

            have_warning = False
            for value in table["Property"]:
                if len(value) > 0 and value[-1] == "*":
                    have_warning = True
                    break

            have_acf_warning = False
            for value in table["tau"]:
                if len(value) > 0 and value[-1] == "^":
                    have_warning = True
                    break

            if have_warning or have_acf_warning:
                printer.normal("\n")
            if have_warning:
                printer.normal(
                    __(
                        "          * this property has less than 100 independent "
                        "samples, so may not be accurate.",
                        wrap=False,
                        dedent=False,
                    )
                )

            if have_acf_warning:
                printer.normal(
                    __(
                        "          ^ there are not enough samples after "
                        "equilibration to plot the ACF.",
                        wrap=False,
                        dedent=False,
                    )
                )

        # Process the output to get the energy
        lines = iter(output)
        for line in lines:
            if "   Step" in line:
                tmp = line.split()
                try:
                    i = tmp.index("TotEng")
                except ValueError:
                    pass
                else:
                    line = next(lines)
                    tmp = line.split()
                    energy = float(tmp[i])
                    data["energy"] = energy
                    data["energy,units"] = "kcal/mol"

                # Check for reaxff enthalpy offset
                if self.parent.ff_form() == "reaxff":
                    Eat = self.parent._atomic_energy_sum
                    if Eat != 0.0:
                        dHf = data["energy"] + Eat
                        data["DfH0_reax"] = dHf
                    data["energy,units"] = "kcal/mol"
            if line.startswith("Loop time of"):
                try:
                    tmp = line.split()
                    _time = round(float(tmp[3]), 2)
                    _procs = int(tmp[5])
                    _steps = int(tmp[8])
                    _natoms = int(tmp[11])
                    _type = self._calculation
                    self.results[f"t_{_type}"] = _time
                    self.results[f"np_{_type}"] = _procs
                    self.results[f"steps_{_type}"] = _steps
                    self.results[f"natoms_{_type}"] = _natoms
                except Exception as _e:
                    print(f"LAMMPS loop time: {_e}")

        # See if forces have been dumped
        wdir = Path(self.directory)
        path = wdir / "forces.dump"
        if path.exists():
            tmp = self.parent.get_dump(path)

            fields = tmp["fields"]
            if "fx" in fields and "fy" in fields and "fz" in fields:
                ix = fields.index("fx")
                iy = fields.index("fy")
                iz = fields.index("fz")
                fxs = tmp["data"][-1][ix]
                fys = tmp["data"][-1][iy]
                fzs = tmp["data"][-1][iz]
                gradients = [[-fx, -fy, -fz] for fx, fy, fz in zip(fxs, fys, fzs)]
                data["gradients"] = gradients
                data["gradients,units"] = "kcal/mol/angstrom"

        # Get the configuration
        _, configuration = self.get_system_configuration(None)

        # Add static properties such as density for e.g NVE and NVT calculations
        if "density" not in data and configuration.periodicity == 3:
            data["density"] = configuration.density
        if "a" not in data and configuration.periodicity == 3:
            data["a"] = configuration.cell.a
        if "b" not in data and configuration.periodicity == 3:
            data["b"] = configuration.cell.b
        if "c" not in data and configuration.periodicity == 3:
            data["c"] = configuration.cell.c
        if "volume" not in data and configuration.periodicity == 3:
            data["volume"] = configuration.cell.volume

        # Put any requested results into variables or tables
        self.store_results(
            configuration=configuration,
            data=data | self.results,
            create_tables=self.parameters["create tables"].get(),
        )

    def get_input(self, extras=None):
        """Get the input for an energy calculation for LAMMPS"""

        P = self.parameters.values_to_dict()

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        self.description = []
        self.description.append(__(self.description_text(PP), **PP, indent=4 * " "))

        lines = []

        filename = f"@{self._id[-1]}+forces.dump"
        lines.append("")
        lines.append(f"# {self.header}")
        lines.append("")
        lines.append(f"dump                1 all custom 1 {filename} id fx fy fz")
        lines.append("run                 0")
        lines.append("undump              1")

        return {
            "script": lines,
            "postscript": None,
            "use python": False,
        }



# -*- coding: utf-8 -*-

"""The graphical part of a LAMMPS minimization step"""

import lammps_step
import seamm_widgets as sw
import tkinter.ttk as ttk


class TkMinimization(lammps_step.TkEnergy):
    def __init__(
        self, tk_flowchart=None, node=None, canvas=None, x=None, y=None, w=200, h=50
    ):
        """Initialize a node

        Keyword arguments:
        """

        super().__init__(
            tk_flowchart=tk_flowchart, node=node, canvas=canvas, x=x, y=y, w=w, h=h
        )

    def create_dialog(self):
        """Create the dialog!"""

        # Let parent classes do their thing.
        frame = super().create_dialog(title="Edit Minimization step")

        # Shortcut for parameters
        P = self.node.parameters

        # Frame to isolate widgets
        opt_frame = self["optimization"] = ttk.LabelFrame(
            frame,
            borderwidth=4,
            relief="sunken",
            text="Geometry Optimization",
            labelanchor="n",
            padding=10,
        )

        for key in (
            "convergence",
            "etol",
            "ftol",
            "nsteps",
            "nevaluations",
            "minimizer",
            "timestep",
            "optimize cell",
        ):
            self[key] = P[key].widget(opt_frame)

        # and binding to change as needed
        for key in ("optimize cell",):
            self[key].bind("<<ComboboxSelected>>", self.reset_dialog)
            self[key].bind("<Return>", self.reset_dialog)
            self[key].bind("<FocusOut>", self.reset_dialog)
        for key in ("convergence", "minimizer"):
            self[key].bind("<<ComboboxSelected>>", self.reset_optimization)
            self[key].bind("<Return>", self.reset_optimization)
            self[key].bind("<FocusOut>", self.reset_optimization)

        # Pressure/stress controls -- create frame for them
        p_frame = self["pressure frame"] = ttk.LabelFrame(
            self["frame"],
            borderwidth=4,
            relief="sunken",
            text="Pressure",
            labelanchor="n",
            padding=10,
        )

        # Create the widgets
        for key in (
            "system type",
            "P",
            "allow shear",
            "use_stress",
            "couple",
            "nreset",
        ):
            self[key] = P[key].widget(p_frame)

        # The stress/pressure section is quite complicated due to
        # couplings, etc. so put in its own subsection
        s_frame = self["stress_frame"] = ttk.Frame(
            p_frame, borderwidth=4, relief="sunken"
        )

        # The last widgets
        for key in (
            "Sxx",
            "Syy",
            "Szz",
            "Sxy",
            "Sxz",
            "Syz",
        ):
            self[key] = P[key].widget(s_frame, unitswidth=6)
            self[key].units.bind("<<ComboboxSelected>>", self._handle_units)

        # The binding above and below is so that we can keep the units of stress and
        # pressure consistent throughout.
        self["P"].units.bind("<<ComboboxSelected>>", self._handle_units)

        # and labels for the directions, couplings, etc.
        for text in (
            "XX",
            "YY",
            "ZZ",
            "YZ",
            "XZ",
            "XY",
            "XX+YY",
            "XX+ZZ",
            "YY+ZZ",
            "XX+YY+ZZ",
        ):
            self[text] = ttk.Label(s_frame, text=text)

        self["stress"] = ttk.Label(s_frame, text="Stress:")

        # and adding bindings for appropriate widgets
        for key in ("system type", "use_stress"):
            self[key].bind("<<ComboboxSelected>>", self.reset_pressure_frame)
        for key in ("couple", "allow shear"):
            self[key].bind("<<ComboboxSelected>>", self.reset_stress_frame)

        # Top level needs to call reset_dialog
        if self.node.calculation == "minimization":
            self.reset_dialog()

        return frame

    def _handle_units(self, event=None):
        """Callback used to keep the pressure units consistent.

        When the units of a stress or the pressure is changed, sets all the others to
        the same units. This is needed because most of the units are hidden.

        Parameters
        ----------
        event : tkinter event object
        """

        units = event.widget.get()

        for key in (
            "P",
            "Sxx",
            "Syy",
            "Szz",
            "Sxy",
            "Sxz",
            "Syz",
        ):
            self[key].units.set(units)

    def reset_dialog(self, widget=None):
        """Layout the widgets as needed for the current state"""
        optimize_cell = self["optimize cell"].get() != "no"

        frame = self["frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        row = 0
        if optimize_cell:
            keys = ("optimization", "pressure frame", "structure")
        else:
            keys = ("optimization", "structure")
        for key in keys:
            self[key].grid(row=row, column=0, sticky="new")
            row += 1

        self.reset_optimization()
        self.reset_pressure_frame()

        return row

    def reset_optimization(self, widget=None):
        convergence = self["convergence"].get()
        minimizer = self["minimizer"].get()

        frame = self["optimization"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        widgets = []
        widgets2 = []
        row = 0

        for key in ("minimizer", "convergence"):
            self[key].grid(row=row, column=0, columnspan=2, sticky="ew")
            widgets.append(self[key])
            row += 1

        if convergence == "custom":
            for key in ("etol", "ftol"):
                self[key].grid(row=row, column=1, sticky="ew")
                widgets2.append(self[key])
                row += 1

        for key in ("nsteps", "nevaluations"):
            self[key].grid(row=row, column=0, columnspan=2, sticky="ew")
            widgets.append(self[key])
            row += 1

        if minimizer in ("Fire", "QuickMin"):
            for key in ("timestep",):
                self[key].grid(row=row, column=0, columnspan=2, sticky="ew")
                widgets.append(self[key])
                row += 1

        for key in ("optimize cell",):
            self[key].grid(row=row, column=0, columnspan=2, sticky="ew")
            widgets.append(self[key])
            row += 1

        width1 = sw.align_labels(widgets, sticky="e")
        width2 = sw.align_labels(widgets2, sticky="e")
        frame.columnconfigure(0, minsize=width1 - width2 + 30)

    def reset_pressure_frame(self, widget=None):
        """Layout the widgets for the pressure/stress control
        as needed for the current state"""

        # Get the values that control the layout
        system_type = self["system type"].get()
        use_stress = "pressure" not in self["use_stress"].get()

        # Remove all the current widgets
        p_frame = self["pressure frame"]
        for slave in p_frame.grid_slaves():
            slave.grid_forget()

        row = 0
        widgets = []
        # and place the needed ones back in
        self["system type"].grid(row=row, column=0, sticky="w")
        widgets.append(self["system type"])
        row += 1

        if system_type == "fluid":
            self["P"].grid(row=row, column=0, sticky="w")
            widgets.append(self["P"])
            row += 1
        else:
            for key in ("use_stress", "allow shear", "couple"):
                self[key].grid(row=row, column=0, sticky="w")
                widgets.append(self[key])
                row += 1

            if use_stress:
                self["stress_frame"].grid(row=row, column=0, sticky="w")
                row += 1
                # and lay out the stress terms
                self.reset_stress_frame()
            else:
                self["P"].grid(row=row, column=0, sticky="w")
                widgets.append(self["P"])
                row += 1

        self["nreset"].grid(row=row, column=0, sticky="w")
        widgets.append(self["nreset"])
        row += 1

        sw.align_labels(widgets, sticky="e")

    def reset_stress_frame(self, widget=None):
        """Layout the widgets for the pressure/stress
        as needed for the current state.

        We use labels across the top and left side of the
        table of stresses, then hide the labels and units
        of all the entries, except for the last entry in the
        row, which displays the units too.

        It is a bit tricky getting the column labels to lign
        up for the last item in the row because it is wider
        with the units displaying. The weird stuff with the
        columnspan=3 and setting the 'uniform' attribute of
        the columns does this. There might be better ways, but
        this works ... just be careful changing it :-).
        """

        # Get the values that control the layout
        couple = self["couple"].get()
        allow_shear = self["allow shear"].get() != "no"

        # Remove all the current widgets
        frame = self["stress_frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        frame.columnconfigure(1, weight=0)
        frame.columnconfigure(2, weight=0)
        frame.columnconfigure(3, weight=0)
        frame.columnconfigure(4, weight=0)
        frame.columnconfigure(5, weight=0)
        frame.columnconfigure(6, weight=0)

        row = 0
        # and place the needed ones back in
        if couple == "x, y and z":
            # all stresses
            self["XX+YY+ZZ"].grid(row=row, column=1)
            if allow_shear:
                self["YZ"].grid(row=row, column=2)
                self["XZ"].grid(row=row, column=3)
                self["XY"].grid(row=row, column=4)
            row += 1
            self["stress"].grid(row=row, column=0, sticky="e")
            self["Sxx"].grid(row=row, column=1, sticky="ew")
            if not allow_shear:
                self["Sxx"].show("combobox", "units")
                frame.columnconfigure(1, weight=1, uniform="b")
            else:
                self["Sxx"].show("combobox")
                self["Syz"].grid(row=row, column=2, sticky="ew")
                self["Sxz"].grid(row=row, column=3, sticky="ew")
                self["Sxy"].grid(row=row, column=4, sticky="ew")
                self["Syz"].show("combobox")
                self["Sxz"].show("combobox")
                self["Sxy"].show("combobox", "units")
                frame.columnconfigure(1, weight=1)
                frame.columnconfigure(2, weight=1)
                frame.columnconfigure(3, weight=1)
                frame.columnconfigure(4, weight=1)
            row += 1
        elif couple == "x and y":
            # couple xx and yy
            self["XX+YY"].grid(row=row, column=1)
            self["ZZ"].grid(row=row, column=2)
            if allow_shear:
                self["YZ"].grid(row=row, column=3)
                self["XZ"].grid(row=row, column=4)
                self["XY"].grid(row=row, column=5)
            row += 1
            self["stress"].grid(row=row, column=0, sticky="e")
            self["Sxx"].grid(row=row, column=1, sticky="ew")
            self["Sxx"].show("combobox")
            self["Szz"].grid(row=row, column=2, sticky="ew")
            if not allow_shear:
                self["Szz"].show("combobox", "units")
                frame.columnconfigure(1, weight=1)
                frame.columnconfigure(2, weight=1)
            else:
                self["Szz"].show("combobox")
                self["Syz"].grid(row=row, column=3, sticky="ew")
                self["Sxz"].grid(row=row, column=4, sticky="ew")
                self["Sxy"].grid(row=row, column=5, sticky="ew")
                self["Syz"].show("combobox")
                self["Sxz"].show("combobox")
                self["Sxy"].show("combobox", "units")
                frame.columnconfigure(1, weight=1)
                frame.columnconfigure(2, weight=1)
                frame.columnconfigure(3, weight=1)
                frame.columnconfigure(4, weight=1)
                frame.columnconfigure(5, weight=1)
            row += 1
        elif couple == "x and z":
            # couple xx and zz
            self["XX+ZZ"].grid(row=row, column=1)
            self["YY"].grid(row=row, column=2)
            if allow_shear:
                self["YZ"].grid(row=row, column=3)
                self["XZ"].grid(row=row, column=4)
                self["XY"].grid(row=row, column=5)
            row += 1
            self["stress"].grid(row=row, column=0, sticky="e")
            self["Sxx"].grid(row=row, column=1, sticky="ew")
            self["Syy"].grid(row=row, column=2, sticky="ew")
            self["Sxx"].show("combobox")
            if not allow_shear:
                self["Syy"].show("combobox", "units")
                frame.columnconfigure(1, weight=1)
                frame.columnconfigure(2, weight=1)
            else:
                self["Syy"].show("combobox")
                self["Syz"].grid(row=row, column=3, sticky="ew")
                self["Sxz"].grid(row=row, column=4, sticky="ew")
                self["Sxy"].grid(row=row, column=5, sticky="ew")
                self["Syz"].show("combobox")
                self["Sxz"].show("combobox", "units")
                self["Sxy"].show("combobox")
                frame.columnconfigure(1, weight=1)
                frame.columnconfigure(2, weight=1)
                frame.columnconfigure(3, weight=1)
                frame.columnconfigure(4, weight=1)
                frame.columnconfigure(5, weight=1)
            row += 1
        elif couple == "y and z":
            # couple yy and zz
            self["XX"].grid(row=row, column=1)
            self["YY+ZZ"].grid(row=row, column=2)
            if allow_shear:
                self["YZ"].grid(row=row, column=3)
                self["XZ"].grid(row=row, column=4)
                self["XY"].grid(row=row, column=5)
            row += 1
            self["stress"].grid(row=row, column=0, sticky="e")
            self["Sxx"].grid(row=row, column=1, sticky="ew")
            self["Syy"].grid(row=row, column=2, sticky="ew")
            self["Sxx"].show("combobox")
            if not allow_shear:
                self["Syy"].show("combobox", "units")
                frame.columnconfigure(1, weight=1)
                frame.columnconfigure(2, weight=1)
            else:
                self["Syy"].show("combobox")
                self["Syz"].grid(row=row, column=3, sticky="ew")
                self["Sxz"].grid(row=row, column=4, sticky="ew")
                self["Sxy"].grid(row=row, column=5, sticky="ew")
                self["Syz"].show("combobox")
                self["Sxz"].show("combobox")
                self["Sxy"].show("combobox", "units")
                frame.columnconfigure(1, weight=1)
                frame.columnconfigure(2, weight=1)
                frame.columnconfigure(3, weight=1)
                frame.columnconfigure(4, weight=1)
                frame.columnconfigure(5, weight=1)
            row += 1
        else:
            # if couple == 'none':
            # all stresses
            self["XX"].grid(row=row, column=1)
            self["YY"].grid(row=row, column=2)
            self["ZZ"].grid(row=row, column=3)
            if allow_shear:
                self["YZ"].grid(row=row, column=4)
                self["XZ"].grid(row=row, column=5)
                self["XY"].grid(row=row, column=6)
            row += 1
            self["stress"].grid(row=row, column=0, sticky="e")
            self["Sxx"].grid(row=row, column=1, sticky="ew")
            self["Sxx"].show("combobox")
            self["Syy"].grid(row=row, column=2, sticky="ew")
            self["Syy"].show("combobox")
            self["Szz"].grid(row=row, column=3, sticky="ew")
            if not allow_shear:
                self["Szz"].show("combobox", "units")
                frame.columnconfigure(1, weight=1)
                frame.columnconfigure(2, weight=1)
                frame.columnconfigure(3, weight=1)
            else:
                self["Szz"].show("combobox")
                self["Syz"].grid(row=row, column=4, sticky="ew")
                self["Sxz"].grid(row=row, column=5, sticky="ew")
                self["Sxy"].grid(row=row, column=6, sticky="ew")
                self["Syz"].show("combobox")
                self["Sxz"].show("combobox")
                self["Sxy"].show("combobox", "units")
                frame.columnconfigure(1, weight=1)
                frame.columnconfigure(2, weight=1)
                frame.columnconfigure(3, weight=1)
                frame.columnconfigure(4, weight=1)
                frame.columnconfigure(5, weight=1)
                frame.columnconfigure(6, weight=1)
            row += 1



# -*- coding: utf-8 -*-

"""Control parameters for minimization"""
import logging

from .energy_parameters import EnergyParameters


logger = logging.getLogger(__name__)


class MinimizationParameters(EnergyParameters):
    """The control parameters for minimization in LAMMPS"""

    parameters = {
        "convergence": {
            "default": "normal",
            "kind": "string",
            "format_string": "s",
            "enumeration": (
                "crude",
                "loose",
                "normal",
                "tight",
                "custom",
            ),
            "description": "Convergence:",
            "help_text": "The convergence criterion for the minimization.",
        },
        "etol": {
            "default": 0.0,
            "kind": "float",
            "default_units": "kcal/mol",
            "format_string": ".2f",
            "description": "Energy change:",
            "help_text": "Converged when the change in energy is less than this.",
        },
        "ftol": {
            "default": 0.1,
            "kind": "float",
            "default_units": "kcal/mol/Å",
            "format_string": ".2f",
            "description": "Force convergence:",
            "help_text": "Converged when the norm of the forces is less than this.",
        },
        "nsteps": {
            "default": "30*nAtoms",
            "kind": "string",
            "format_string": "",
            "description": "Maximum steps:",
            "help_text": "The maximum number of steps.",
        },
        "nevaluations": {
            "default": "3*nSteps",
            "kind": "string",
            "format_string": "",
            "description": "Maximum evaluations:",
            "help_text": "The maximum number of energy evaluations.",
        },
        "minimizer": {
            "default": "Conjugate Gradient",
            "kind": "string",
            "format_string": "",
            "enumeration": (
                "Conjugate Gradient",
                "Steepest Descent",
                "Hessian-Free truncated Newton",
                "QuickMin",
                "Fire",
            ),
            "description": "Minimizer:",
            "help_text": "Which minimizer to use",
        },
        "timestep": {
            "default": "10",
            "kind": "float",
            "default_units": "fs",
            "enumeration": ("normal", "accurate but slow", "coarse but fast"),
            "format_string": ".1f",
            "description": "Timestep:",
            "help_text": (
                "The time step for the numerical integration in "
                "the dynamics. About 10x that used in MD works well."
            ),
        },
        "optimize cell": {
            "default": "yes",
            "kind": "boolean",
            "format_string": "s",
            "enumeration": ("no", "yes"),
            "description": "Optimize cell",
            "help_text": "Whether to optimize the unit cell.",
        },
        "system type": {
            "default": "fluid",
            "kind": "string",
            "format_string": "s",
            "enumeration": ("fluid", "solid"),
            "description": "Type of system:",
            "help_text": (
                "Whether the system is a fluid, "
                "which flows under shear stress, "
                "or a solid, which can resist "
                "small shear stresses."
            ),
        },
        "allow shear": {
            "default": "no",
            "kind": "boolean",
            "format_string": "s",
            "enumeration": ("no", "yes"),
            "description": "Allow the cell to shear",
            "help_text": "Whether the cell angles can change.",
        },
        "use_stress": {
            "default": "isotropic pressure",
            "kind": "enumeration",
            "enumeration": ("isotropic pressure", "general stress"),
            "format_string": "s",
            "description": "Apply",
            "help_text": (
                "Specify whether to apply (isotropic) "
                "pressure to the system or explicit stresses "
                "for each different direction."
            ),
        },
        "couple": {
            "default": "x, y and z",
            "kind": "enumeration",
            "enumeration": ("x, y and z", "x and y", "x and z", "y and z", "none"),
            "format_string": "s",
            "description": "Directions to couple:",
            "help_text": (
                "The stress in these directions will be "
                "averaged and the cell dilated in fixed "
                "proportions in these directions."
            ),
        },
        "P": {
            "default": 1.0,
            "kind": "float",
            "default_units": "atm",
            "format_string": ".2f",
            "enumeration": tuple(),
            "description": "Pressure:",
            "help_text": "The applied pressure.",
        },
        "Sxx": {
            "default": -1.0,
            "kind": "float",
            "enumeration": ("fixed",),
            "default_units": "atm",
            "format_string": "s",
            "description": "Sxx:",
            "help_text": "The components of the stress tensor.",
        },
        "Syy": {
            "default": -1.0,
            "kind": "float",
            "enumeration": ("fixed",),
            "default_units": "atm",
            "format_string": "s",
            "description": "Syy:",
            "help_text": "The components of the stress tensor.",
        },
        "Szz": {
            "default": -1.0,
            "kind": "float",
            "enumeration": ("fixed",),
            "default_units": "atm",
            "format_string": "s",
            "description": "Szz:",
            "help_text": "The components of the stress tensor.",
        },
        "Syz": {
            "default": "fixed",
            "kind": "float",
            "enumeration": ("fixed",),
            "default_units": "atm",
            "format_string": "s",
            "description": "Syz:",
            "help_text": "The components of the stress tensor.",
        },
        "Sxz": {
            "default": "fixed",
            "kind": "float",
            "enumeration": ("fixed",),
            "default_units": "atm",
            "format_string": "s",
            "description": "Sxz:",
            "help_text": "The components of the stress tensor.",
        },
        "Sxy": {
            "default": "fixed",
            "kind": "float",
            "enumeration": ("fixed",),
            "default_units": "atm",
            "format_string": "s",
            "description": "Sxy:",
            "help_text": "The components of the stress tensor.",
        },
        "nreset": {
            "default": "never",
            "kind": "integer",
            "default_units": None,
            "format_string": "d",
            "enumeration": ("never",),
            "description": "Frequency to reset reference cell:",
            "help_text": (
                "How often, in number of steps, to reset "
                "the reference cell for the strain energy."
            ),
        },
    }

    def __init__(self, defaults={}, data=None):
        """Initialize the instance, by default from the default
        parameters given in the class"""

        super().__init__(
            defaults={
                **MinimizationParameters.parameters,
                **defaults,
            },
            data=data,
        )



# -*- coding: utf-8 -*-

"""The graphical part of a LAMMPS Initialization step"""

import logging
import tkinter as tk
import tkinter.ttk as ttk

import seamm
import seamm_widgets as sw

logger = logging.getLogger(__name__)


class TkInitialization(seamm.TkNode):
    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=None,
        y=None,
        w=200,
        h=50,
        my_logger=logger,
    ):
        """Setup  the LAMMPS initialization node.

        Keyword arguments:
        """

        # # Set the logging level for this module if requested
        # if 'lammps_tk_initialization_log_level' in self.options:
        #     logger.setLevel(self.options.lammps_tk_initialization_log_level)
        #     logger.critical(
        #         'Set log level to {}'.format(
        #             self.options.lammps_tk_initialization_log_level
        #         )
        #     )

        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h,
            my_logger=my_logger,
        )

    def right_click(self, event):
        """Probably need to add our dialog..."""

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def create_dialog(self):
        """Create the dialog!"""
        frame = super().create_dialog("Edit LAMMPS Initialization Step")

        # Shortcut for parameters
        P = self.node.parameters

        # Frame for options for all systems
        general = self["general"] = ttk.LabelFrame(
            frame,
            text="For all systems",
            relief=tk.SUNKEN,
            borderwidth=5,
            labelanchor=tk.N,
        )

        for key in (
            "cutoff",
            "shift_nonbond",
            "atomic charges",
            "qeq convergence",
            "qeq iterations",
        ):
            self[key] = P[key].widget(general)

        # Frame for the periodic system options, i.e. kspace, etc.
        periodic = self["periodic"] = ttk.LabelFrame(
            frame,
            text="For periodic systems",
            relief=tk.SUNKEN,
            borderwidth=5,
            labelanchor=tk.N,
        )

        for key in (
            "tail_correction",
            "kspace_method",
            "kspace_accuracy",
            "kspace_smallq",
        ):
            self[key] = P[key].widget(periodic)

        self["atomic charges"].bind("<<ComboboxSelected>>", self.general_cb)
        self["kspace_method"].bind("<<ComboboxSelected>>", self.kspace_method_cb)

        # Grid in the static part of the dialog

        self["general"].grid(row=0, column=0, sticky=tk.EW, pady=10)

        self["periodic"].grid(row=1, column=0, sticky=tk.EW, pady=10)
        self.general_cb()
        self.kspace_method_cb()

    def general_cb(self, event=None):
        """Grid the widgets into the dialog, depending on the current values
        of key variables. This provides a dyamic presentation to the user.
        """
        frame = self["general"]
        charge_method = self["atomic charges"].get()

        # Remove any widgets previously packed
        for slave in frame.grid_slaves():
            slave.grid_forget()

        row = 0
        widgets = []
        widgets2 = []
        for key in ("cutoff", "shift_nonbond", "atomic charges"):
            self[key].grid(row=row, column=0, columnspan=2, sticky=tk.EW)
            widgets.append(self[key])
            row += 1

        if charge_method.startswith("charge "):
            for key in ("qeq convergence", "qeq iterations"):
                self[key].grid(row=row, column=1, sticky=tk.EW)
                widgets2.append(self[key])
                row += 1

        w1 = sw.align_labels(widgets, sticky=tk.E)
        if len(widgets2) > 0:
            w2 = sw.align_labels(widgets2, sticky=tk.E)
            frame.columnconfigure(0, minsize=w1 - w2 + 30)

    def kspace_method_cb(self, event=None):
        """Grid the widgets into the dialog, depending on the current values
        of key variables. This provides a dyamic presentation to the user.
        """

        # Remove any widgets previously packed
        for slave in self["periodic"].grid_slaves():
            slave.grid_forget()

        row = 0
        self["kspace_method"].grid(row=row, column=0, sticky=tk.W)
        row += 1

        method = self["kspace_method"].get()
        if method != "none":
            self["kspace_accuracy"].grid(row=row, column=0, sticky=tk.W)
            row += 1

            if "few charged" in method or self.is_expr(method[0]):
                self["kspace_smallq"].grid(row=row, column=0, sticky=tk.W)
                row += 1

        if self.is_expr(method) or "dispersion" not in method:
            self["tail_correction"].grid(row=row, column=0, sticky=tk.W)
            row += 1



# -*- coding: utf-8 -*-

"""
lammps_step
A step for running molecular dynamics using LAMMPS.
"""

# Bring up the classes so that they appear to be directly in
# the lammps_step package.


# The metadata
from .metadata import metadata

from .lammps_units import set_lammps_unit_system
from .lammps_units import get_lammps_unit_system
from .lammps_units import to_lammps_units
from .lammps_units import from_lammps_units
from .lammps_units import lammps_units

from .lammps import bond_style
from .lammps import angle_style
from .lammps import dihedral_style
from .lammps import improper_style

from .lammps import LAMMPS
from .lammps_step import LAMMPSStep
from .tk_lammps import TkLAMMPS

from .custom import Custom
from .custom_parameters import CustomParameters
from .custom_step import CustomStep
from .tk_custom import TkCustom

from .initialization import Initialization
from .initialization_parameters import InitializationParameters
from .initialization_parameters import kspace_methods, charge_methods
from .initialization_step import InitializationStep
from .tk_initialization import TkInitialization

from .energy import Energy
from .energy_parameters import EnergyParameters
from .energy_step import EnergyStep
from .tk_energy import TkEnergy

from .minimization import Minimization
from .minimization_parameters import MinimizationParameters
from .minimization_step import MinimizationStep
from .tk_minimization import TkMinimization

from .velocities import Velocities
from .velocities_parameters import VelocitiesParameters
from .velocities_step import VelocitiesStep
from .tk_velocities import TkVelocities

from .nve import NVE
from .nve_parameters import NVE_Parameters
from .nve_step import NVEStep
from .tk_nve import TkNVE

from .nvt import NVT
from .nvt import thermostat_metadata
from .nvt_parameters import NVT_Parameters
from .nvt_step import NVTStep
from .tk_nvt import TkNVT

from .npt import NPT
from .npt_parameters import NPT_Parameters
from .npt_step import NPTStep
from .tk_npt import TkNPT

from .heat_flux_step import HeatFluxStep
from .heat_flux import HeatFlux
from .heat_flux_parameters import HeatFluxParameters
from .tk_heat_flux import TkHeatFlux

# Handle versioneer
from ._version import get_versions

__author__ = """Paul Saxe"""
__email__ = "psaxe@molssi.org"
versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `lammps_step` package, lammps_units module."""

import lammps_step

# Need pint for formats.
from seamm_util import ureg, Q_, units_class  # noqa: F401


def test_default_system():
    """Testing that we can convert to/from the default unit system."""
    value = "2.0 kJ/mol"
    converted = lammps_step.to_lammps_units(value)
    assert (
        f"{value} in real units is {converted:.4}"
        == "2.0 kJ/mol in real units is 0.478"
    )
    assert (
        f'{converted:.4} --> {lammps_step.from_lammps_units(converted, "kJ/mol"):~.4}'  # noqa: E501
        == "0.478 --> 2.0 kJ / mol"
    )


def test_metal_system():
    """Testing that we can convert to/from the metal unit system."""
    lammps_step.set_lammps_unit_system("metal")
    value = "2.0 kJ/mol"
    converted = lammps_step.to_lammps_units(value)
    assert (
        f"{value} in metal units is {converted:.4}"
        == "2.0 kJ/mol in metal units is 0.02073"
    )
    assert (
        f'{converted:.4} --> {lammps_step.from_lammps_units(converted, "kJ/mol"):~.4}'  # noqa: E501
        == "0.02073 --> 2.0 kJ / mol"
    )



# -*- coding: utf-8 -*-

"""The graphical part of a LAMMPS NVT dynamics step"""

import lammps_step
import logging
import seamm_widgets as sw
import pprint  # nopep8
import tkinter as tk
import tkinter.ttk as ttk

logger = logging.getLogger(__name__)


class TkNVT(lammps_step.TkNVE):
    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=None,
        y=None,
        w=200,
        h=50,
        my_logger=logger,
    ):
        """Initialize a node

        Keyword arguments:
        """

        # Set the logging level for this module if requested
        # if 'lammps_tk_nvt_log_level' in self.options:
        #     logger.setLevel(self.options.lammps_tk_nvt_log_level)
        #     logger.critical(
        #         'Set log level to {}'.format(
        #             self.options.lammps_tk_nvt_log_level
        #         )
        #     )

        # Call the constructor for the superclass
        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h,
            my_logger=my_logger,
        )

    def create_dialog(self, title="Edit NVT dynamics parameters"):
        """Create the edit dialog!

        This is reasonably complicated, so a bit of description
        is in order. The superclass NVE creates the dialog
        along with the basic runtime and timestep.

        This method adds a second frame for setting the temperature and
        thermostat (this is the T part of NVT).

        The layout is handled in part by the NVT superclass, which
        handles the temperature frame. Our part is handled by two
        methods:

        * reset_dialog does the general layout of the large blocks
        * reset_temperature_frame handles the layout of the temperature
          section except
        """

        logger.debug("TkNVT.create_dialog")

        # Let parent classes do their thing.
        super().create_dialog(title=title)

        # Shortcut for parameters
        P = self.node.parameters

        logger.debug("Parameters:\n{}".format(pprint.pformat(P.to_dict())))

        # Temperature frame to isolate widgets
        t_frame = self["temperature_frame"] = ttk.LabelFrame(
            self["frame"],
            borderwidth=4,
            relief="sunken",
            text="Temperature",
            labelanchor="n",
            padding=10,
        )

        for key in lammps_step.NVT_Parameters.parameters:
            self[key] = P[key].widget(t_frame)

        # and binding to change as needed
        self["thermostat"].combobox.bind(
            "<<ComboboxSelected>>", self.reset_temperature_frame
        )

    def reset_dialog(self, widget=None):
        """Layout the widgets as needed for the current state"""

        row = super().reset_dialog()

        self["temperature_frame"].grid(row=row, column=0, sticky=tk.EW, pady=10)
        row += 1
        self.reset_temperature_frame()

        # And how to handle the structure
        if self.node.calculation == "nvt":
            self["structure"].grid(row=row, column=0)
            row += 1

        return row

    def reset_temperature_frame(self, widget=None):
        """Layout the widgets in the temperature frame
        as needed for the current state"""

        thermostat = self["thermostat"].get()

        t_frame = self["temperature_frame"]
        for slave in t_frame.grid_slaves():
            slave.grid_forget()

        row = 0

        # Main controls
        self["T0"].grid(row=row, column=0, columnspan=2, sticky=tk.W)
        row += 1

        self["T1"].grid(row=row, column=0, columnspan=2, sticky=tk.W)
        row += 1

        self["thermostat"].grid(row=row, column=0, columnspan=2, sticky=tk.W)
        row += 1

        sw.align_labels((self["T0"], self["T1"], self["thermostat"]), sticky=tk.E)

        # and controls for specific thermostats
        if thermostat != "velocity rescaling":
            self["Tdamp"].grid(row=row, column=1, sticky=tk.W)
            row += 1

        if thermostat == "Nose-Hoover":
            self["drag"].grid(row=row, column=1, sticky=tk.W)
            row += 1

            self["Tchain"].grid(row=row, column=1, sticky=tk.W)
            row += 1

            self["Tloop"].grid(row=row, column=1, sticky=tk.W)
            row += 1

            sw.align_labels(
                (self["Tdamp"], self["Tchain"], self["Tloop"], self["drag"]),
                sticky=tk.E,
            )
        elif thermostat == "Berendsen":
            pass
        elif "csvr" in thermostat or "csld" in thermostat:
            self["seed"].grid(row=row, column=1, sticky=tk.W)
            row += 1
        elif thermostat == "velocity rescaling":
            self["frequency"].grid(row=row, column=1, sticky=tk.W)
            row += 1

            self["window"].grid(row=row, column=1, sticky=tk.W)
            row += 1

            self["fraction"].grid(row=row, column=1, sticky=tk.W)
            row += 1

            sw.align_labels(
                (self["frequency"], self["window"], self["fraction"]), sticky=tk.E
            )
        elif thermostat == "Langevin":
            self["seed"].grid(row=row, column=1, sticky=tk.W)
            row += 1
            sw.align_labels((self["Tdamp"], self["seed"]), sticky=tk.E)
        else:
            raise RuntimeError(
                "Don't recognize thermostat " + "'{}'".format(thermostat)
            )

        t_frame.columnconfigure(0, minsize=50)

    def handle_dialog(self, result):
        """Handle the user cancelling, closing or OK'ing the dialog.

        If the user cancels or closes the dialog with the x button,
        revert the widgets to their previous state.

        If the user clocks 'OK', save the changes to both the control
        parameters and the results requested.
        """
        if result == "OK":
            # Shortcut for parameters
            P = self.node.parameters

            thermostat = self["thermostat"].get()

            P["T0"].set(self["T0"].get())
            P["T1"].set(self["T1"].get())
            P["thermostat"].set(thermostat)

            if thermostat != "velocity rescaling":
                P["Tdamp"].set(self["Tdamp"].get())

            if thermostat == "Nose-Hoover":
                P["Tchain"].set(self["Tchain"].get())
                P["Tloop"].set(self["Tloop"].get())
                P["drag"].set(self["drag"].get())
            elif (
                "csvr" in thermostat or "csld" in thermostat or thermostat == "Langevin"
            ):
                P["seed"].set(self["seed"].get())
            elif thermostat == "velocity rescaling":
                P["frequency"].set(self["frequency"].get())
                P["window"].set(self["window"].get())
                P["fraction"].set(self["fraction"].get())

        # Let base classes reap their parameters
        super().handle_dialog(result)



# -*- coding: utf-8 -*-

"""The graphical part of a Heat Flux step"""

import tkinter as tk

import lammps_step
import seamm_widgets as sw


class TkHeatFlux(lammps_step.TkNVE):
    """
    The graphical part of a Heat Flux step in a flowchart.

    Attributes
    ----------
    tk_flowchart : TkFlowchart = None
        The flowchart that we belong to.
    node : Node = None
        The corresponding node of the non-graphical flowchart
    canvas: tkCanvas = None
        The Tk Canvas to draw on
    dialog : Dialog
        The Pmw dialog object
    x : int = None
        The x-coordinate of the center of the picture of the node
    y : int = None
        The y-coordinate of the center of the picture of the node
    w : int = 200
        The width in pixels of the picture of the node
    h : int = 50
        The height in pixels of the picture of the node
    self[widget] : dict
        A dictionary of tk widgets built using the information
        contained in Heat Flux_parameters.py

    See Also
    --------
    HeatFlux, TkHeatFlux,
    HeatFluxParameters,
    """

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=None,
        y=None,
        w=200,
        h=50,
    ):
        """
        Initialize a graphical node.

        Parameters
        ----------
        tk_flowchart: Tk_Flowchart
            The graphical flowchart that we are in.
        node: Node
            The non-graphical node for this step.
        namespace: str
            The stevedore namespace for finding sub-nodes.
        canvas: Canvas
           The Tk canvas to draw on.
        x: float
            The x position of the nodes center on the canvas.
        y: float
            The y position of the nodes cetner on the canvas.
        w: float
            The nodes graphical width, in pixels.
        h: float
            The nodes graphical height, in pixels.

        Returns
        -------
        None
        """
        self.dialog = None

        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h,
        )

    def create_dialog(self):
        """
        Create the dialog. A set of widgets will be chosen by default
        based on what is specified in the Heat Flux_parameters
        module.

        Parameters
        ----------
        None

        Returns
        -------
        None

        See Also
        --------
        TkHeatFlux.reset_dialog
        """

        frame = super().create_dialog(title="Heat Flux")

        # Shortcut for parameters
        P = self.node.parameters

        # Then create the widgets
        for key in ("time", "timestep", "heat flux", "use centroid stress", "sampling"):
            self[key] = P[key].widget(frame)

        # and lay them out
        self.reset_dialog()

    def reset_dialog(self, widget=None):
        """Layout the widgets in the dialog.

        The widgets are chosen by default from the information in
        Heat Flux_parameter.

        This function simply lays them out row by row with
        aligned labels. You may wish a more complicated layout that
        is controlled by values of some of the control parameters.
        If so, edit or override this method

        Parameters
        ----------
        widget : Tk Widget = None

        Returns
        -------
        None

        See Also
        --------
        TkHeatFlux.create_dialog
        """

        # Remove any widgets previously packed
        frame = self["frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        # Shortcut for parameters
        P = self.node.parameters

        # keep track of the row in a variable, so that the layout is flexible
        # if e.g. rows are skipped to control such as "method" here
        row = 0
        widgets = []
        for key in ("time", "timestep", "heat flux", "use centroid stress", "sampling"):
            self[key].grid(row=row, column=0, sticky=tk.EW)
            widgets.append(self[key])
            row += 1

        # Align the labels
        sw.align_labels(widgets, sticky=tk.E)

        # And how to handle the structure
        if self.node.calculation == "Heat Flux":
            self["structure"].grid(row=row, column=0)
            row += 1

        # Setup the results if there are any
        have_results = (
            "results" in self.node.metadata and len(self.node.metadata["results"]) > 0
        )
        if have_results and "results" in P:
            self.setup_results()

    def right_click(self, event):
        """
        Handles the right click event on the node.

        Parameters
        ----------
        event : Tk Event

        Returns
        -------
        None

        See Also
        --------
        TkHeatFlux.edit
        """

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)



"""This file contains metadata to help describe the results of LAMMPS
calculations, etc.
"""

metadata = {}

"""Properties that LAMMPS produces, depending on the type of calculation.
"""

metadata["results"] = {
    "energy": {
        "calculation": ["energy", "minimization"],
        "description": "The total energy",
        "dimensionality": "scalar",
        "property": "total energy#LAMMPS#{model}",
        "type": "float",
        "units": "kcal/mol",
    },
    "gradients": {
        "calculation": ["energy"],
        "description": "The gradients",
        "dimensionality": "[3, n_atoms]",
        "type": "float",
        "units": "kcal/mol/Å",
    },
    "model": {
        "description": "The model used",
        "dimensionality": "scalar",
        "type": "string",
    },
    "DfH0_reax": {
        "calculation": ["energy", "minimization"],
        "description": "the Reax appriximation to the enthalpy of formation",
        "dimensionality": "scalar",
        "property": "DfH0_reax#LAMMPS#{model}",
        "type": "float",
        "units": "kcal/mol",
        "format": ".2f",
    },
    "T": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "temperature",
        "dimensionality": "scalar",
        "property": "temperature#LAMMPS#{model}",
        "type": "float",
        "units": "K",
    },
    "T,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of temperature",
        "dimensionality": "scalar",
        "property": "temperature, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "K",
    },
    "T,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of temperature",
        "dimensionality": "scalar",
        "property": "temperature, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "T,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of temperature sampling",
        "dimensionality": "scalar",
        "property": "temperature, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "P": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "pressure",
        "dimensionality": "scalar",
        "property": "pressure#LAMMPS#{model}",
        "type": "float",
        "units": "atm",
    },
    "P,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of pressure",
        "dimensionality": "scalar",
        "property": "pressure, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "atm",
    },
    "P,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of pressure",
        "dimensionality": "scalar",
        "property": "pressure, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "P,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of pressure sampling",
        "dimensionality": "scalar",
        "property": "pressure, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "Sxx": {
        "description": "Sxx",
        "dimensionality": "scalar",
        "property": "Sxx#LAMMPS#{model}",
        "type": "float",
        "units": "atm",
    },
    "Sxx,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of Sxx",
        "dimensionality": "scalar",
        "property": "Sxx, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "atm",
    },
    "Sxx,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of Sxx",
        "dimensionality": "scalar",
        "property": "Sxx, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "Sxx,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of Sxx sampling",
        "dimensionality": "scalar",
        "property": "Sxx, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "Syy": {
        "description": "Syy",
        "dimensionality": "scalar",
        "property": "Syy#LAMMPS#{model}",
        "type": "float",
        "units": "atm",
    },
    "Syy,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of Syy",
        "dimensionality": "scalar",
        "property": "Syy, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "atm",
    },
    "Syy,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of Syy",
        "dimensionality": "scalar",
        "property": "Syy, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "Syy,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of Syy sampling",
        "dimensionality": "scalar",
        "property": "Syy, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "Szz": {
        "description": "Szz",
        "dimensionality": "scalar",
        "property": "Szz#LAMMPS#{model}",
        "type": "float",
        "units": "atm",
    },
    "Szz,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of Szz",
        "dimensionality": "scalar",
        "property": "Szz, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "atm",
    },
    "Szz,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of Szz",
        "dimensionality": "scalar",
        "property": "Szz, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "Szz,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of Szz sampling",
        "dimensionality": "scalar",
        "property": "Szz, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "Sxy": {
        "description": "Sxy",
        "dimensionality": "scalar",
        "property": "Sxy#LAMMPS#{model}",
        "type": "float",
        "units": "atm",
    },
    "Sxy,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of Sxy",
        "dimensionality": "scalar",
        "property": "Sxy, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "atm",
    },
    "Sxy,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of Sxy",
        "dimensionality": "scalar",
        "property": "Sxy, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "Sxy,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of Sxy sampling",
        "dimensionality": "scalar",
        "property": "Sxy, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "Sxz": {
        "description": "Sxz",
        "dimensionality": "scalar",
        "property": "Sxz#LAMMPS#{model}",
        "type": "float",
        "units": "atm",
    },
    "Sxz,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of Sxz",
        "dimensionality": "scalar",
        "property": "Sxz, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "atm",
    },
    "Sxz,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of Sxz",
        "dimensionality": "scalar",
        "property": "Sxz, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "Sxz,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of Sxz sampling",
        "dimensionality": "scalar",
        "property": "Sxz, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "Syz": {
        "description": "Syz",
        "dimensionality": "scalar",
        "property": "Syz#LAMMPS#{model}",
        "type": "float",
        "units": "atm",
    },
    "Syz,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of Syz",
        "dimensionality": "scalar",
        "property": "Syz, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "atm",
    },
    "Syz,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of Syz",
        "dimensionality": "scalar",
        "property": "Syz, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "Syz,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of Syz sampling",
        "dimensionality": "scalar",
        "property": "Syz, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "density": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "density",
        "dimensionality": "scalar",
        "property": "density#LAMMPS#{model}",
        "type": "float",
        "units": "g/ml",
    },
    "density,stderr": {
        "calculation": ["npt"],
        "description": "stderr of density",
        "dimensionality": "scalar",
        "property": "density, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "g/ml",
    },
    "density,tau": {
        "calculation": ["npt"],
        "description": "autocorrelation time of density",
        "dimensionality": "scalar",
        "property": "density, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "density,inefficiency": {
        "calculation": ["npt"],
        "description": "statistical inefficiency of density sampling",
        "dimensionality": "scalar",
        "property": "density, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "V": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "V",
        "dimensionality": "scalar",
        "property": "volume#LAMMPS#{model}",
        "type": "float",
        "units": "g/ml",
    },
    "V,stderr": {
        "calculation": ["npt"],
        "description": "stderr of V",
        "dimensionality": "scalar",
        "property": "volume, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "g/ml",
    },
    "V,tau": {
        "calculation": ["npt"],
        "description": "autocorrelation time of V",
        "dimensionality": "scalar",
        "property": "volume, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "V,inefficiency": {
        "calculation": ["npt"],
        "description": "statistical inefficiency of V sampling",
        "dimensionality": "scalar",
        "property": "volume, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "a": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "cell parameter 'a'",
        "dimensionality": "scalar",
        "property": "cell_a#LAMMPS#{model}",
        "type": "float",
        "units": "Å",
    },
    "a,stderr": {
        "calculation": ["npt"],
        "description": "stderr of cell 'a'",
        "dimensionality": "scalar",
        "property": "cell_a, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "Å",
    },
    "a,tau": {
        "calculation": ["npt"],
        "description": "autocorrelation time of cell 'a'",
        "dimensionality": "scalar",
        "property": "cell_a, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "a,inefficiency": {
        "calculation": ["npt"],
        "description": "statistical inefficiency of cell 'a' sampling",
        "dimensionality": "scalar",
        "property": "cell_a, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "b": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "cell parameter 'b'",
        "dimensionality": "scalar",
        "property": "cell_b#LAMMPS#{model}",
        "type": "float",
        "units": "Å",
    },
    "b,stderr": {
        "calculation": ["npt"],
        "description": "stderr of cell 'b'",
        "dimensionality": "scalar",
        "property": "cell_b, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "Å",
    },
    "b,tau": {
        "calculation": ["npt"],
        "description": "autocorrelation time of cell 'b'",
        "dimensionality": "scalar",
        "property": "cell_b, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "b,inefficiency": {
        "calculation": ["npt"],
        "description": "statistical inefficiency of cell 'b' sampling",
        "dimensionality": "scalar",
        "property": "cell_b, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "c": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "cell parameter 'c'",
        "dimensionality": "scalar",
        "property": "cell_c#LAMMPS#{model}",
        "type": "float",
        "units": "Å",
    },
    "c,stderr": {
        "calculation": ["npt"],
        "description": "stderr of cell 'c'",
        "dimensionality": "scalar",
        "property": "cell_c, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "Å",
    },
    "c,tau": {
        "calculation": ["npt"],
        "description": "autocorrelation time of cell 'c'",
        "dimensionality": "scalar",
        "property": "cell_c, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "c,inefficiency": {
        "calculation": ["npt"],
        "description": "statistical inefficiency of cell 'c' sampling",
        "dimensionality": "scalar",
        "property": "cell_c, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "alpha": {
        "calculation": ["npt"],
        "description": "cell parameter 'alpha'",
        "dimensionality": "scalar",
        "property": "cell_alpha#LAMMPS#{model}",
        "type": "float",
        "units": "degree",
    },
    "alpha,stderr": {
        "calculation": ["npt"],
        "description": "stderr of cell 'alpha'",
        "dimensionality": "scalar",
        "property": "cell_alpha, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "degree",
    },
    "alpha,tau": {
        "calculation": ["npt"],
        "description": "autocorrelation time of cell 'alpha'",
        "dimensionality": "scalar",
        "property": "cell_alpha, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "alpha,inefficiency": {
        "calculation": ["npt"],
        "description": "statistical inefficiency of cell 'alpha' sampling",
        "dimensionality": "scalar",
        "property": "cell_alpha, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "beta": {
        "calculation": ["npt"],
        "description": "cell parameter 'beta'",
        "dimensionality": "scalar",
        "property": "cell_beta#LAMMPS#{model}",
        "type": "float",
        "units": "degree",
    },
    "beta,stderr": {
        "calculation": ["npt"],
        "description": "stderr of cell 'beta'",
        "dimensionality": "scalar",
        "property": "cell_beta, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "degree",
    },
    "beta,tau": {
        "calculation": ["npt"],
        "description": "autocorrelation time of cell 'beta'",
        "dimensionality": "scalar",
        "property": "cell_beta, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "beta,inefficiency": {
        "calculation": ["npt"],
        "description": "statistical inefficiency of cell 'beta' sampling",
        "dimensionality": "scalar",
        "property": "cell_beta, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "gamma": {
        "calculation": ["npt"],
        "description": "cell parameter 'gamma'",
        "dimensionality": "scalar",
        "property": "cell_gamma#LAMMPS#{model}",
        "type": "float",
        "units": "degree",
    },
    "gamma,stderr": {
        "calculation": ["npt"],
        "description": "stderr of cell 'gamma'",
        "dimensionality": "scalar",
        "property": "cell_gamma, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "degree",
    },
    "gamma,tau": {
        "calculation": ["npt"],
        "description": "autocorrelation time of cell 'gamma'",
        "dimensionality": "scalar",
        "property": "cell_gamma, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "gamma,inefficiency": {
        "calculation": ["npt"],
        "description": "statistical inefficiency of cell 'gamma' sampling",
        "dimensionality": "scalar",
        "property": "cell_gamma, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "Etot": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "total energy",
        "dimensionality": "scalar",
        "property": "total energy#LAMMPS#{model}",
        "type": "float",
        "units": "kcal/mol",
    },
    "Etot,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of total energy",
        "dimensionality": "scalar",
        "property": "total energy, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "kcal/mol",
    },
    "Etot,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of total energy",
        "dimensionality": "scalar",
        "property": "total energy, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "Etot,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of total energy sampling",
        "dimensionality": "scalar",
        "property": "total energy, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "Eke": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "kinetic energy",
        "dimensionality": "scalar",
        "property": "kinetic energy#LAMMPS#{model}",
        "type": "float",
        "units": "kcal/mol",
    },
    "Eke,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of kinetic energy",
        "dimensionality": "scalar",
        "property": "kinetic energy, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "kcal/mol",
    },
    "Eke,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of kinetic energy",
        "dimensionality": "scalar",
        "property": "kinetic energy, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "Eke,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of kinetic energy sampling",
        "dimensionality": "scalar",
        "property": "kinetic energy, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "Epe": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "potential energy",
        "dimensionality": "scalar",
        "property": "potential energy#LAMMPS#{model}",
        "type": "float",
        "units": "kcal/mol",
    },
    "Epe,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of potential energy",
        "dimensionality": "scalar",
        "property": "potential energy, stderr#LAMMPS#{model}",
        "type": "float",
        "units": "kcal/mol",
    },
    "Epe,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of potential energy",
        "dimensionality": "scalar",
        "property": "potential energy, tau#LAMMPS#{model}",
        "type": "float",
        "units": "fs",
    },
    "Epe,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of potential energy sampling",
        "dimensionality": "scalar",
        "property": "potential energy, inefficiency#LAMMPS#{model}",
        "type": "float",
        "units": "",
    },
    "Epair": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "nonbonded (vdW & electrostatic) energy",
        "dimensionality": "scalar",
        "type": "float",
        "units": "kcal/mol",
    },
    "Epair,stderr": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "stderr of nonbond energy",
        "dimensionality": "scalar",
        "type": "float",
        "units": "kcal/mol",
    },
    "Epair,tau": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "autocorrelation time of nonbond energy",
        "dimensionality": "scalar",
        "type": "float",
        "units": "fs",
    },
    "Epair,inefficiency": {
        "calculation": [
            "nve",
            "nvt",
            "npt",
        ],
        "description": "statistical inefficiency of nonbond energy sampling",
        "dimensionality": "scalar",
        "type": "float",
        "units": "",
    },
    # Timings
    "t_lammps_wall": {
        "description": "The wall clock time for LAMMPS",
        "dimensionality": "scalar",
        "type": "float",
        "units": "s",
        "format": ".1f",
    },
    "SEAMM elapsed time": {
        "description": "total elapsed time for LAMMPS",
        "dimensionality": "scalar",
        "type": "float",
        "units": "s",
        "format": ".1f",
    },
    "SEAMM np": {
        "description": "number of processors for LAMMPS",
        "dimensionality": "scalar",
        "type": "integer",
        "format": "d",
    },
    "t_nve": {
        "calculation": ["nve"],
        "description": "The time for the NVE step",
        "dimensionality": "scalar",
        "type": "float",
        "units": "s",
    },
    "np_nve": {
        "calculation": ["nve"],
        "description": "The number of processors for the NVE step",
        "dimensionality": "scalar",
        "type": "integer",
        "units": "s",
    },
    "nsteps_nve": {
        "calculation": ["nve"],
        "description": "The number of steps in the NVE step",
        "dimensionality": "scalar",
        "type": "integer",
        "units": "s",
    },
    "natoms_nve": {
        "calculation": ["nve"],
        "description": "The number of atoms in the NVE step",
        "dimensionality": "scalar",
        "type": "integer",
        "units": "s",
    },
    "t_nvt": {
        "calculation": ["nvt"],
        "description": "The time for the NVT step",
        "dimensionality": "scalar",
        "type": "float",
        "units": "s",
    },
    "np_nvt": {
        "calculation": ["nvt"],
        "description": "The number of processors for the NVT step",
        "dimensionality": "scalar",
        "type": "integer",
        "units": "s",
    },
    "nsteps_nvt": {
        "calculation": ["nvt"],
        "description": "The number of steps in the NVT step",
        "dimensionality": "scalar",
        "type": "integer",
        "units": "s",
    },
    "natoms_nvt": {
        "calculation": ["nvt"],
        "description": "The number of atoms in the NVT step",
        "dimensionality": "scalar",
        "type": "integer",
        "units": "s",
    },
    "t_npt": {
        "calculation": ["npt"],
        "description": "The time for the NPT step",
        "dimensionality": "scalar",
        "type": "float",
        "units": "s",
    },
    "np_npt": {
        "calculation": ["npt"],
        "description": "The number of processors for the NPT step",
        "dimensionality": "scalar",
        "type": "integer",
        "units": "s",
    },
    "nsteps_npt": {
        "calculation": ["npt"],
        "description": "The number of steps in the NPT step",
        "dimensionality": "scalar",
        "type": "integer",
        "units": "s",
    },
    "natoms_npt": {
        "calculation": ["npt"],
        "description": "The number of atoms in the NPT step",
        "dimensionality": "scalar",
        "type": "integer",
        "units": "s",
    },
    "minimizer": {
        "calculation": ["minimization"],
        "description": "the minimizer used",
        "dimensionality": "scalar",
        "type": "string",
    },
    "optimization is converged": {
        "calculation": ["minimization"],
        "description": "whether the geometry optimization converged",
        "dimensionality": "scalar",
        "type": "boolean",
        "format": "s",
    },
    "N steps optimization": {
        "calculation": ["minimization"],
        "description": "number of optimization steps",
        "dimensionality": "scalar",
        "type": "integer",
    },
    "N force evaluations": {
        "calculation": ["minimization"],
        "description": "number of force evaluations",
        "dimensionality": "scalar",
        "type": "integer",
    },
    "energy change": {
        "calculation": ["minimization"],
        "description": "the last energy change",
        "dimensionality": "scalar",
        "type": "float",
        "units": "kcal/mol",
        "format": ".2f",
    },
    "energy change threshold": {
        "calculation": ["minimization"],
        "description": "the threshold for the energy change",
        "dimensionality": "scalar",
        "type": "float",
        "units": "kcal/mol/Å",
        "format": ".2f",
    },
    "force norm": {
        "calculation": ["minimization"],
        "description": "the l2 norm of the force vector",
        "dimensionality": "scalar",
        "type": "float",
        "units": "kcal/mol/Å",
        "format": ".2f",
    },
    "force norm threshold": {
        "calculation": ["minimization"],
        "description": "the l2 norm of the force vector threshold",
        "dimensionality": "scalar",
        "type": "float",
        "units": "kcal/mol/Å",
        "format": ".2f",
    },
    "RMSD": {
        "calculation": ["minimization"],
        "description": "RMSD with H removed",
        "dimensionality": "scalar",
        "type": "float",
        "units": "Å",
    },
    "displaced atom": {
        "calculation": ["minimization"],
        "description": "Atom index with largest displacement",
        "dimensionality": "scalar",
        "type": "int",
    },
    "maximum displacement": {
        "calculation": ["minimization"],
        "description": "Maximum displacement of an atom",
        "dimensionality": "scalar",
        "type": "float",
        "units": "Å",
    },
    "RMSD with H": {
        "calculation": ["minimization"],
        "description": "RMSD including H atoms",
        "dimensionality": "scalar",
        "type": "float",
        "units": "Å",
    },
    "displaced atom with H": {
        "calculation": ["minimization"],
        "description": "Atom index with largest displacement, including H",
        "dimensionality": "scalar",
        "type": "int",
    },
    "maximum displacement with H": {
        "calculation": ["minimization"],
        "description": "Maximum displacement of an atom, including H",
        "dimensionality": "scalar",
        "type": "float",
        "units": "Å",
    },
}



# -*- coding: utf-8 -*-
"""Control parameters for a single-point energy (SPE) in LAMMPS"""

import logging
import seamm

logger = logging.getLogger(__name__)


class EnergyParameters(seamm.Parameters):
    """The control parameters for Energy dynamics in LAMMPS"""

    parameters = {
        "results": {
            "default": {},
            "kind": "dictionary",
            "default_units": None,
            "enumeration": tuple(),
            "format_string": "",
            "description": "results",
            "help_text": ("The results to save to variables or in " "tables. "),
        },
        "create tables": {
            "default": "yes",
            "kind": "boolean",
            "default_units": None,
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Create tables as needed:",
            "help_text": (
                "Whether to create tables as needed for "
                "results being saved into tables."
            ),
        },
    }

    def __init__(self, defaults={}, data=None):
        """Initialize the instance, by default from the default
        parameters given in the class"""

        super().__init__(
            defaults={
                **EnergyParameters.parameters,
                **seamm.standard_parameters.structure_handling_parameters,
                **defaults,
            },
            data=data,
        )

        # Do any local editing of defaults
        tmp = self["system name"]
        tmp._data["enumeration"] = ["simulated with {model}", *tmp.enumeration[1:]]
        tmp.default = "keep current name"

        tmp = self["configuration name"]
        tmp._data["enumeration"] = ["simulated with {model}", *tmp.enumeration]
        tmp.default = "simulated with {model}"
